"""
A submodule to hold classes that inherit from :class:`Node`.

I.e. downstream stuff relative to :class:`Node`.
Includes factory functions for such classes, etc.

:class:`Workflow` is exempted from this directory and kept at the main level due to its
importance as the entry point for users.
"""
