"""
A submodule to hold classes that :class:`Node` (and/or :class:`Channel`) inherit from.

I.e. upstream stuff relative to :class:`Node`.
Each class is intended to provide a well-defined interface to some functionality.
Ideally, the mixins should be lightweight and not interfere with each other.
"""
